"""Support functions for testing UWS code."""

from __future__ import annotations

from datetime import timedelta
from typing import Annotated, Self

from arq.connections import RedisSettings
from fastapi import Form, Query
from pydantic import BaseModel, SecretStr

from safir.arq import ArqMode
from safir.uws import ParametersModel, UWSConfig, UWSJobParameter, UWSRoute

__all__ = [
    "SimpleParameters",
    "build_uws_config",
]


class SimpleWorkerParameters(BaseModel):
    name: str


class SimpleParameters(ParametersModel[SimpleWorkerParameters]):
    name: str

    @classmethod
    def from_job_parameters(cls, params: list[UWSJobParameter]) -> Self:
        assert len(params) == 1
        assert params[0].parameter_id == "name"
        return cls(name=params[0].value)

    def to_worker_parameters(self) -> SimpleWorkerParameters:
        return SimpleWorkerParameters(name=self.name)


async def _get_dependency(
    name: Annotated[str, Query()],
) -> list[UWSJobParameter]:
    return [UWSJobParameter(parameter_id="name", value=name)]


async def _post_dependency(
    name: Annotated[str, Form()],
) -> list[UWSJobParameter]:
    return [UWSJobParameter(parameter_id="name", value=name)]


def build_uws_config(database_url: str, database_password: str) -> UWSConfig:
    """Set up a test configuration."""
    return UWSConfig(
        arq_mode=ArqMode.test,
        arq_redis_settings=RedisSettings(host="localhost", port=6379),
        async_post_route=UWSRoute(
            dependency=_post_dependency, summary="Create async job"
        ),
        database_url=database_url,
        database_password=SecretStr(database_password),
        execution_duration=timedelta(minutes=10),
        lifetime=timedelta(days=1),
        parameters_type=SimpleParameters,
        signing_service_account="signer@example.com",
        slack_webhook=SecretStr("https://example.com/fake-webhook"),
        sync_get_route=UWSRoute(
            dependency=_get_dependency, summary="Sync request"
        ),
        sync_post_route=UWSRoute(
            dependency=_post_dependency, summary="Sync request"
        ),
        worker="hello",
    )



"""Test fixtures."""

from __future__ import annotations

from collections.abc import AsyncIterator, Iterator
from datetime import timedelta

import pytest
import pytest_asyncio
import respx
from redis.asyncio import Redis
from testcontainers.postgres import PostgresContainer
from testcontainers.redis import RedisContainer

from safir.testing.gcs import MockStorageClient, patch_google_storage
from safir.testing.kubernetes import MockKubernetesApi, patch_kubernetes
from safir.testing.slack import MockSlackWebhook, mock_slack_webhook


@pytest.fixture(scope="session")
def database_password() -> str:
    return "INSECURE@%PASSWORD/"


@pytest.fixture(scope="session")
def database_url(database_password: str) -> Iterator[str]:
    """Start a PostgreSQL database and return a URL for it."""
    with PostgresContainer(
        driver="asyncpg", username="safir", password=database_password
    ) as postgres:
        yield postgres.get_connection_url()


@pytest.fixture
def mock_gcs() -> Iterator[MockStorageClient]:
    yield from patch_google_storage(
        expected_expiration=timedelta(hours=1), bucket_name="some-bucket"
    )


@pytest.fixture
def mock_kubernetes() -> Iterator[MockKubernetesApi]:
    yield from patch_kubernetes()


@pytest.fixture
def mock_slack(respx_mock: respx.Router) -> MockSlackWebhook:
    return mock_slack_webhook("https://example.com/slack", respx_mock)


@pytest.fixture(scope="session")
def redis() -> Iterator[RedisContainer]:
    """Start a Redis container."""
    with RedisContainer() as redis:
        yield redis


@pytest_asyncio.fixture
async def redis_client(redis: RedisContainer) -> AsyncIterator[Redis]:
    """Create a Redis client for testing.

    This must be done separately for each test since it's tied to the per-test
    event loop, and therefore must be separated from the session-shared Redis
    server container.
    """
    host = redis.get_container_host_ip()
    port = redis.get_exposed_port(6379)
    client = Redis(host=host, port=port, db=0)
    yield client
    await client.aclose()



"""Tests for errors from the UWS API."""

from __future__ import annotations

from dataclasses import dataclass

import pytest
from httpx import AsyncClient

from safir.testing.slack import MockSlackWebhook
from safir.uws import UWSJobParameter
from safir.uws._dependencies import UWSFactory


@dataclass
class PostTest:
    """Encapsulates the data a test POST."""

    url: str
    data: dict[str, str]


@pytest.mark.asyncio
async def test_errors(
    client: AsyncClient, uws_factory: UWSFactory, mock_slack: MockSlackWebhook
) -> None:
    job_service = uws_factory.create_job_service()
    await job_service.create(
        "user",
        run_id="some-run-id",
        params=[UWSJobParameter(parameter_id="name", value="June")],
    )

    # No user specified.
    routes = [
        "/test/jobs/1",
        "/test/jobs/1/destruction",
        "/test/jobs/1/error",
        "/test/jobs/1/executionduration",
        "/test/jobs/1/owner",
        "/test/jobs/1/parameters",
        "/test/jobs/1/phase",
        "/test/jobs/1/quote",
        "/test/jobs/1/results",
    ]
    for route in routes:
        r = await client.get(route)
        assert r.status_code == 422
        assert r.text.startswith("UsageError")

    # Wrong user specified.
    for route in routes:
        r = await client.get(
            route, headers={"X-Auth-Request-User": "otheruser"}
        )
        assert r.status_code == 403
        assert r.text.startswith("AuthorizationError")

    # Job does not exist.
    for route in (r.replace("/1", "/2") for r in routes):
        r = await client.get(route, headers={"X-Auth-Request-User": "user"})
        assert r.status_code == 404
        assert r.text.startswith("UsageError")

    # Check no user specified with POST routes.
    tests = [
        PostTest("/test/jobs/1", {"action": "DELETE"}),
        PostTest(
            "/test/jobs/1/destruction", {"destruction": "2021-09-10T10:01:02Z"}
        ),
        PostTest(
            "/test/jobs/1/executionduration", {"executionduration": "1200"}
        ),
        PostTest("/test/jobs/1/phase", {"phase": "RUN"}),
    ]
    for test in tests:
        r = await client.post(test.url, data=test.data)
        assert r.status_code == 422
        assert r.text.startswith("UsageError")

    # Wrong user specified.
    for test in tests:
        r = await client.post(
            test.url,
            data=test.data,
            headers={"X-Auth-Request-User": "otheruser"},
        )
        assert r.status_code == 403
        assert r.text.startswith("AuthorizationError")

    # Job does not exist.
    for test in tests:
        url = test.url.replace("/1", "/2")
        r = await client.post(
            url, data=test.data, headers={"X-Auth-Request-User": "user"}
        )
        assert r.status_code == 404
        assert r.text.startswith("UsageError")

    # Finally, test all the same things with the one supported DELETE.
    r = await client.delete("/test/jobs/1")
    assert r.status_code == 422
    assert r.text.startswith("UsageError")
    r = await client.delete(
        "/test/jobs/1", headers={"X-Auth-Request-User": "otheruser"}
    )
    assert r.status_code == 403
    assert r.text.startswith("AuthorizationError")
    r = await client.delete(
        "/test/jobs/2", headers={"X-Auth-Request-User": "user"}
    )
    assert r.status_code == 404
    assert r.text.startswith("UsageError")

    # Try some bogus destruction and execution duration parameters.
    tests = [
        PostTest("/test/jobs/1/destruction", {"destruction": "next tuesday"}),
        PostTest("/test/jobs/1/destruction", {"DESTruction": "next tuesday"}),
        PostTest(
            "/test/jobs/1/destruction", {"destruction": "2021-09-10T10:01:02"}
        ),
        PostTest(
            "/test/jobs/1/destruction", {"destrucTION": "2021-09-10T10:01:02"}
        ),
        PostTest("/test/jobs/1/executionduration", {"executionduration": "0"}),
        PostTest("/test/jobs/1/executionduration", {"executionDUration": "0"}),
        PostTest(
            "/test/jobs/1/executionduration", {"executionduration": "-1"}
        ),
        PostTest(
            "/test/jobs/1/executionduration", {"executionDUration": "-1"}
        ),
        PostTest(
            "/test/jobs/1/executionduration", {"executionduration": "fred"}
        ),
        PostTest(
            "/test/jobs/1/executionduration", {"executionDUration": "fred"}
        ),
    ]
    for test in tests:
        r = await client.post(
            test.url, data=test.data, headers={"X-Auth-Request-User": "user"}
        )
        assert r.status_code == 422
        assert r.text.startswith("UsageError")

    # Test bogus phase for async job creation.
    r = await client.post(
        "/test/jobs?phase=START",
        headers={"X-Auth-Request-User": "user"},
        data={"runid": "some-run-id", "name": "Jane"},
    )
    assert r.status_code == 422
    r = await client.post(
        "/test/jobs",
        headers={"X-Auth-Request-User": "user"},
        data={"runid": "some-run-id", "name": "Jane", "phase": "START"},
    )
    assert r.status_code == 422

    # None of these errors should have produced Slack errors.
    assert mock_slack.messages == []



"""UWS schema

Revision ID: e9299566bc19
Revises:
Create Date: 2024-09-11 00:00:57.336783+00:00

"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "e9299566bc19"
down_revision: str | None = None
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "job",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("message_id", sa.String(length=64), nullable=True),
        sa.Column("owner", sa.String(length=64), nullable=False),
        sa.Column(
            "phase",
            sa.Enum(
                "PENDING",
                "QUEUED",
                "EXECUTING",
                "COMPLETED",
                "ERROR",
                "UNKNOWN",
                "HELD",
                "SUSPENDED",
                "ABORTED",
                "ARCHIVED",
                name="executionphase",
            ),
            nullable=False,
        ),
        sa.Column("run_id", sa.String(length=64), nullable=True),
        sa.Column("creation_time", sa.DateTime(), nullable=False),
        sa.Column("start_time", sa.DateTime(), nullable=True),
        sa.Column("end_time", sa.DateTime(), nullable=True),
        sa.Column("destruction_time", sa.DateTime(), nullable=False),
        sa.Column("execution_duration", sa.Integer(), nullable=False),
        sa.Column("quote", sa.DateTime(), nullable=True),
        sa.Column(
            "error_type",
            sa.Enum("TRANSIENT", "FATAL", name="errortype"),
            nullable=True,
        ),
        sa.Column(
            "error_code",
            sa.Enum(
                "AUTHENTICATION_ERROR",
                "AUTHORIZATION_ERROR",
                "MULTIVALUED_PARAM_NOT_SUPPORTED",
                "ERROR",
                "SERVICE_UNAVAILABLE",
                "USAGE_ERROR",
                name="errorcode",
            ),
            nullable=True,
        ),
        sa.Column("error_message", sa.Text(), nullable=True),
        sa.Column("error_detail", sa.Text(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        "by_owner_phase",
        "job",
        ["owner", "phase", "creation_time"],
        unique=False,
    )
    op.create_index(
        "by_owner_time", "job", ["owner", "creation_time"], unique=False
    )
    op.create_table(
        "job_parameter",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("job_id", sa.Integer(), nullable=False),
        sa.Column("parameter", sa.String(length=64), nullable=False),
        sa.Column("value", sa.Text(), nullable=False),
        sa.Column("is_post", sa.Boolean(), nullable=False),
        sa.ForeignKeyConstraint(["job_id"], ["job.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        "by_parameter", "job_parameter", ["job_id", "parameter"], unique=False
    )
    op.create_table(
        "job_result",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("job_id", sa.Integer(), nullable=False),
        sa.Column("result_id", sa.String(length=64), nullable=False),
        sa.Column("sequence", sa.Integer(), nullable=False),
        sa.Column("url", sa.String(length=256), nullable=False),
        sa.Column("size", sa.Integer(), nullable=True),
        sa.Column("mime_type", sa.String(length=64), nullable=True),
        sa.ForeignKeyConstraint(["job_id"], ["job.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        "by_result_id", "job_result", ["job_id", "result_id"], unique=True
    )
    op.create_index(
        "by_sequence", "job_result", ["job_id", "sequence"], unique=True
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("by_sequence", table_name="job_result")
    op.drop_index("by_result_id", table_name="job_result")
    op.drop_table("job_result")
    op.drop_index("by_parameter", table_name="job_parameter")
    op.drop_table("job_parameter")
    op.drop_index("by_owner_time", table_name="job")
    op.drop_index("by_owner_phase", table_name="job")
    op.drop_table("job")
    # ### end Alembic commands ###



"""SQLAlchemy schema for the UWS database."""

from __future__ import annotations

from datetime import datetime

from sqlalchemy import ForeignKey, Index, String, Text
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship
from vo_models.uws.types import ErrorType, ExecutionPhase

from ._models import ErrorCode

__all__ = [
    "Job",
    "JobParameter",
    "JobResult",
    "UWSSchemaBase",
]


class UWSSchemaBase(DeclarativeBase):
    """SQLAlchemy declarative base for the UWS database schema."""


class JobResult(UWSSchemaBase):
    """Table holding job results."""

    __tablename__ = "job_result"

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    job_id: Mapped[int] = mapped_column(
        ForeignKey("job.id", ondelete="CASCADE")
    )
    result_id: Mapped[str] = mapped_column(String(64))
    sequence: Mapped[int]
    url: Mapped[str] = mapped_column(String(256))
    size: Mapped[int | None]
    mime_type: Mapped[str | None] = mapped_column(String(64))

    __table_args__ = (
        Index("by_sequence", "job_id", "sequence", unique=True),
        Index("by_result_id", "job_id", "result_id", unique=True),
    )


class JobParameter(UWSSchemaBase):
    """Table holding parameters to UWS jobs."""

    __tablename__ = "job_parameter"

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    job_id: Mapped[int] = mapped_column(
        ForeignKey("job.id", ondelete="CASCADE")
    )
    parameter: Mapped[str] = mapped_column(String(64))
    value: Mapped[str] = mapped_column(Text)
    is_post: Mapped[bool] = mapped_column(default=False)

    __table_args__ = (Index("by_parameter", "job_id", "parameter"),)


class Job(UWSSchemaBase):
    """Table holding UWS jobs.

    Notes
    -----
    The details of how the relationships are defined are chosen to allow this
    schema to be used with async SQLAlchemy. Review the SQLAlchemy asyncio
    documentation carefully before making changes. There are a lot of
    surprises and sharp edges.
    """

    __tablename__ = "job"

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    message_id: Mapped[str | None] = mapped_column(String(64))
    owner: Mapped[str] = mapped_column(String(64))
    phase: Mapped[ExecutionPhase]
    run_id: Mapped[str | None] = mapped_column(String(64))
    creation_time: Mapped[datetime]
    start_time: Mapped[datetime | None]
    end_time: Mapped[datetime | None]
    destruction_time: Mapped[datetime]
    execution_duration: Mapped[int]
    quote: Mapped[datetime | None]
    error_type: Mapped[ErrorType | None]
    error_code: Mapped[ErrorCode | None]
    error_message: Mapped[str | None] = mapped_column(Text)
    error_detail: Mapped[str | None] = mapped_column(Text)

    parameters: Mapped[list[JobParameter]] = relationship(
        cascade="delete", lazy="selectin", uselist=True
    )
    results: Mapped[list[JobResult]] = relationship(
        cascade="delete", lazy="selectin", uselist=True
    )

    __table_args__ = (
        Index("by_owner_phase", "owner", "phase", "creation_time"),
        Index("by_owner_time", "owner", "creation_time"),
    )



"""Return internal objects as XML responses."""

from __future__ import annotations

import jinja2
from fastapi import Request, Response
from fastapi.templating import Jinja2Templates

from safir.datetime import isodatetime

from ._models import UWSJob, UWSJobError
from ._results import ResultStore

__all__ = ["UWSTemplates"]

_environment = jinja2.Environment(
    loader=jinja2.PackageLoader("safir", "uws/templates"),
    undefined=jinja2.StrictUndefined,
    autoescape=True,
)
_environment.filters["isodatetime"] = isodatetime
_templates = Jinja2Templates(env=_environment)


class UWSTemplates:
    """Template responses for the UWS protocol.

    This also includes VOSI-Availability since it was convenient to provide.
    """

    def __init__(self, result_store: ResultStore) -> None:
        self._result_store = result_store

    def error(self, request: Request, error: UWSJobError) -> Response:
        """Return the error of a job as an XML response."""
        return _templates.TemplateResponse(
            request,
            "error.xml",
            {"error": error},
            media_type="application/xml",
        )

    async def job(self, request: Request, job: UWSJob) -> Response:
        """Return a job as an XML response."""
        results = [self._result_store.sign_url(r) for r in job.results]
        return _templates.TemplateResponse(
            request,
            "job.xml",
            {"job": job, "results": results},
            media_type="application/xml",
        )

    def parameters(self, request: Request, job: UWSJob) -> Response:
        """Return the parameters for a job as an XML response."""
        return _templates.TemplateResponse(
            request,
            "parameters.xml",
            {"job": job},
            media_type="application/xml",
        )



"""Test handling of jobs that fail."""

from __future__ import annotations

import pytest
from httpx import AsyncClient

from safir.arq.uws import WorkerFatalError, WorkerTransientError
from safir.datetime import isodatetime
from safir.testing.slack import MockSlackWebhook
from safir.testing.uws import MockUWSJobRunner
from safir.uws import UWSJobParameter
from safir.uws._dependencies import UWSFactory
from safir.uws._exceptions import TaskError

ERRORED_JOB = """
<uws:job
    version="1.1"
    xsi:schemaLocation="http://www.ivoa.net/xml/UWS/v1.0 UWS.xsd"
    xmlns:xml="http://www.w3.org/XML/1998/namespace"
    xmlns:uws="http://www.ivoa.net/xml/UWS/v1.0"
    xmlns:xlink="http://www.w3.org/1999/xlink"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
  <uws:jobId>1</uws:jobId>
  <uws:ownerId>user</uws:ownerId>
  <uws:phase>ERROR</uws:phase>
  <uws:creationTime>{}</uws:creationTime>
  <uws:startTime>{}</uws:startTime>
  <uws:endTime>{}</uws:endTime>
  <uws:executionDuration>600</uws:executionDuration>
  <uws:destruction>{}</uws:destruction>
  <uws:parameters>
    <uws:parameter id="name">Sarah</uws:parameter>
  </uws:parameters>
  <uws:errorSummary type="{}" hasDetail="{}">
    <uws:message>{}</uws:message>
  </uws:errorSummary>
</uws:job>
"""

JOB_ERROR_SUMMARY = """
<?xml version="1.0" encoding="UTF-8"?>
<VOTABLE version="1.4" xmlns="http://www.ivoa.net/xml/VOTable/v1.3">
  <INFO name="QUERY_STATUS" value="ERROR">
{}
  </INFO>
</VOTABLE>
"""


@pytest.mark.asyncio
async def test_temporary_error(
    client: AsyncClient,
    runner: MockUWSJobRunner,
    uws_factory: UWSFactory,
    mock_slack: MockSlackWebhook,
) -> None:
    job_service = uws_factory.create_job_service()
    await job_service.create(
        "user", params=[UWSJobParameter(parameter_id="name", value="Sarah")]
    )

    # The pending job has no error.
    r = await client.get(
        "/test/jobs/1/error", headers={"X-Auth-Request-User": "user"}
    )
    assert r.status_code == 404

    # Execute the job.
    r = await client.post(
        "/test/jobs/1/phase",
        headers={"X-Auth-Request-User": "user"},
        data={"PHASE": "RUN"},
    )
    assert r.status_code == 303
    await runner.mark_in_progress("user", "1")
    exc = WorkerTransientError("Something failed")
    result = TaskError.from_worker_error(exc)
    job = await runner.mark_complete("user", "1", result)

    # Check the results.
    assert job.start_time
    assert job.end_time
    r = await client.get(
        "/test/jobs/1", headers={"X-Auth-Request-User": "user"}
    )
    assert r.status_code == 200
    assert r.text == ERRORED_JOB.strip().format(
        isodatetime(job.creation_time),
        isodatetime(job.start_time),
        isodatetime(job.end_time),
        isodatetime(job.destruction_time),
        "transient",
        "false",
        "ServiceUnavailable Something failed",
    )

    # Retrieve the error separately.
    r = await client.get(
        "/test/jobs/1/error", headers={"X-Auth-Request-User": "user"}
    )
    assert r.status_code == 200
    assert r.text == JOB_ERROR_SUMMARY.strip().format(
        "ServiceUnavailable Something failed"
    )

    # For now, this shouldn't have resulted in Slack errors.
    assert mock_slack.messages == []


@pytest.mark.asyncio
async def test_fatal_error(
    client: AsyncClient,
    runner: MockUWSJobRunner,
    uws_factory: UWSFactory,
    mock_slack: MockSlackWebhook,
) -> None:
    job_service = uws_factory.create_job_service()
    await job_service.create(
        "user", params=[UWSJobParameter(parameter_id="name", value="Sarah")]
    )

    # Start the job.
    r = await client.post(
        "/test/jobs/1/phase",
        headers={"X-Auth-Request-User": "user"},
        data={"PHASE": "RUN"},
    )
    assert r.status_code == 303
    await runner.mark_in_progress("user", "1")
    exc = WorkerFatalError("Whoops", "Some details")
    result = TaskError.from_worker_error(exc)
    job = await runner.mark_complete("user", "1", result)

    # Check the results.
    assert job.start_time
    assert job.end_time
    r = await client.get(
        "/test/jobs/1", headers={"X-Auth-Request-User": "user"}
    )
    assert r.status_code == 200
    assert r.text == ERRORED_JOB.strip().format(
        isodatetime(job.creation_time),
        isodatetime(job.start_time),
        isodatetime(job.end_time),
        isodatetime(job.destruction_time),
        "fatal",
        "true",
        "Error Whoops",
    )

    # Retrieve the error separately.
    r = await client.get(
        "/test/jobs/1/error", headers={"X-Auth-Request-User": "user"}
    )
    assert r.status_code == 200
    assert r.text == JOB_ERROR_SUMMARY.strip().format(
        "Error Whoops\n\nSome details"
    )

    # For now, this shouldn't have resulted in Slack errors.
    assert mock_slack.messages == []


@pytest.mark.asyncio
async def test_unknown_error(
    client: AsyncClient,
    runner: MockUWSJobRunner,
    uws_factory: UWSFactory,
    mock_slack: MockSlackWebhook,
) -> None:
    job_service = uws_factory.create_job_service()
    await job_service.create(
        "user", params=[UWSJobParameter(parameter_id="name", value="Sarah")]
    )

    # Start the job.
    r = await client.post(
        "/test/jobs/1/phase",
        headers={"X-Auth-Request-User": "user"},
        data={"PHASE": "RUN"},
    )
    assert r.status_code == 303
    await runner.mark_in_progress("user", "1")
    result = ValueError("Unknown exception")
    job = await runner.mark_complete("user", "1", result)

    # Check the results.
    assert job.start_time
    assert job.end_time
    r = await client.get(
        "/test/jobs/1", headers={"X-Auth-Request-User": "user"}
    )
    assert r.status_code == 200
    assert r.text == ERRORED_JOB.strip().format(
        isodatetime(job.creation_time),
        isodatetime(job.start_time),
        isodatetime(job.end_time),
        isodatetime(job.destruction_time),
        "fatal",
        "true",
        "Error Unknown error executing task",
    )

    # Retrieve the error separately.
    r = await client.get(
        "/test/jobs/1/error", headers={"X-Auth-Request-User": "user"}
    )
    assert r.status_code == 200
    assert r.text == JOB_ERROR_SUMMARY.strip().format(
        "Error Unknown error executing task\n\n"
        "ValueError: Unknown exception"
    )

    # For now, this shouldn't have resulted in Slack errors.
    assert mock_slack.messages == []



"""Test UWS integration with Alembic."""

from __future__ import annotations

from pathlib import Path
from typing import Any

import pytest
from structlog.stdlib import BoundLogger

from safir.uws import DatabaseSchemaError, UWSApplication, UWSConfig


@pytest.mark.asyncio
async def test_database_init(
    uws_config: UWSConfig, logger: BoundLogger
) -> None:
    uws = UWSApplication(uws_config)
    config_path = (
        Path(__file__).parent.parent
        / "data"
        / "database"
        / "uws"
        / "alembic.ini"
    )
    worker_settings = uws.build_worker(
        logger, check_schema=True, alembic_config_path=config_path
    )
    assert worker_settings.on_startup
    assert worker_settings.on_shutdown
    worker_ctx: dict[Any, Any] = {}
    worker_startup = worker_settings.on_startup
    worker_shutdown = worker_settings.on_shutdown

    # Initialize the database without Alembic.
    await uws.initialize_uws_database(logger, reset=True)

    # Initializing a FastAPI app, or creating a UWS worker, should both fail
    # because the database is not current.
    assert not await uws.is_schema_current(logger, config_path)
    with pytest.raises(DatabaseSchemaError):
        await uws.initialize_fastapi(
            logger, check_schema=True, alembic_config_path=config_path
        )
    with pytest.raises(DatabaseSchemaError):
        await worker_startup(worker_ctx)

    # Reinitialize the database with Alembic.
    await uws.initialize_uws_database(
        logger,
        reset=True,
        use_alembic=True,
        alembic_config_path=config_path,
    )

    # Other initializations should now succeed.
    assert await uws.is_schema_current(logger, config_path)
    await uws.initialize_fastapi(
        logger, check_schema=True, alembic_config_path=config_path
    )
    await worker_startup(worker_ctx)

    # Clean up those initializations.
    await uws.shutdown_fastapi()
    await worker_shutdown(worker_ctx)



"""Mock UWS job executor for testing."""

from __future__ import annotations

import asyncio
from datetime import UTC, datetime
from types import TracebackType
from typing import Literal, Self

import structlog
from sqlalchemy.ext.asyncio import AsyncEngine

from safir.arq import JobMetadata, JobResult, MockArqQueue
from safir.database import create_async_session, create_database_engine
from safir.uws import UWSConfig, UWSJob, UWSJobResult
from safir.uws._service import JobService
from safir.uws._storage import JobStore

__all__ = ["MockUWSJobRunner"]


class MockUWSJobRunner:
    """Simulate execution of jobs with a mock queue.

    When running the test suite, the arq queue is replaced with a mock queue
    that doesn't execute workers. That execution has to be simulated by
    manually updating state in the mock queue and running the UWS database
    worker functions that normally would be run automatically by the queue.

    This class wraps that functionality in an async context manager. An
    instance of it is normally provided as a fixture, initialized with the
    same test objects as the test suite.

    Parameters
    ----------
    config
        UWS configuration.
    arq_queue
        Mock arq queue for testing.
    """

    def __init__(self, config: UWSConfig, arq_queue: MockArqQueue) -> None:
        self._config = config
        self._arq = arq_queue
        self._engine: AsyncEngine
        self._store: JobStore
        self._service: JobService

    async def __aenter__(self) -> Self:
        """Create a database session and the underlying service."""
        # This duplicates some of the code in UWSDependency to avoid needing
        # to set up the result store or to expose UWSFactory outside of the
        # Safir package internals.
        self._engine = create_database_engine(
            self._config.database_url,
            self._config.database_password,
            isolation_level="REPEATABLE READ",
        )
        session = await create_async_session(self._engine)
        self._store = JobStore(session)
        self._service = JobService(
            config=self._config,
            arq_queue=self._arq,
            storage=self._store,
            logger=structlog.get_logger("uws"),
        )
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> Literal[False]:
        """Close the database engine and session."""
        await self._engine.dispose()
        return False

    async def get_job_metadata(
        self, username: str, job_id: str
    ) -> JobMetadata:
        """Get the arq job metadata for a job.

        Parameters
        ----------
        job_id
            UWS job ID.

        Returns
        -------
        JobMetadata
            arq job metadata.
        """
        job = await self._service.get(username, job_id)
        assert job.message_id
        return await self._arq.get_job_metadata(job.message_id)

    async def get_job_result(self, username: str, job_id: str) -> JobResult:
        """Get the arq job result for a job.

        Parameters
        ----------
        job_id
            UWS job ID.

        Returns
        -------
        JobMetadata
            arq job metadata.
        """
        job = await self._service.get(username, job_id)
        assert job.message_id
        return await self._arq.get_job_result(job.message_id)

    async def mark_in_progress(
        self, username: str, job_id: str, *, delay: float | None = None
    ) -> UWSJob:
        """Mark a queued job in progress.

        Parameters
        ----------
        username
            Owner of job.
        job_id
            Job ID.
        delay
            How long to delay in seconds before marking the job as complete.

        Returns
        -------
        UWSJob
            Record of the job.
        """
        if delay:
            await asyncio.sleep(delay)
        job = await self._service.get(username, job_id)
        assert job.message_id
        await self._arq.set_in_progress(job.message_id)
        await self._store.mark_executing(job_id, datetime.now(tz=UTC))
        return await self._service.get(username, job_id)

    async def mark_complete(
        self,
        username: str,
        job_id: str,
        results: list[UWSJobResult] | Exception,
        *,
        delay: float | None = None,
    ) -> UWSJob:
        """Mark an in progress job as complete.

        Parameters
        ----------
        username
            Owner of job.
        job_id
            Job ID.
        results
            Results to return. May be an exception to simulate a job failure.
        delay
            How long to delay in seconds before marking the job as complete.

        Returns
        -------
        UWSJob
            Record of the job.
        """
        if delay:
            await asyncio.sleep(delay)
        job = await self._service.get(username, job_id)
        assert job.message_id
        await self._arq.set_complete(job.message_id, result=results)
        job_result = await self._arq.get_job_result(job.message_id)
        await self._store.mark_completed(job_id, job_result)
        return await self._service.get(username, job_id)



"""Storage layer for the UWS implementation."""

from __future__ import annotations

from datetime import datetime, timedelta
from typing import ParamSpec, TypeVar

from sqlalchemy import delete
from sqlalchemy.exc import OperationalError
from sqlalchemy.ext.asyncio import async_scoped_session
from sqlalchemy.future import select
from vo_models.uws.types import ErrorType, ExecutionPhase
from vo_models.vosi.availability import Availability

from safir.arq import JobMetadata, JobResult
from safir.database import (
    datetime_from_db,
    datetime_to_db,
    retry_async_transaction,
)
from safir.datetime import current_datetime

from ._exceptions import TaskError, UnknownJobError
from ._models import (
    ErrorCode,
    UWSJob,
    UWSJobDescription,
    UWSJobError,
    UWSJobParameter,
    UWSJobResult,
)
from ._schema import Job as SQLJob
from ._schema import JobParameter as SQLJobParameter
from ._schema import JobResult as SQLJobResult

T = TypeVar("T")
P = ParamSpec("P")

__all__ = ["JobStore"]


def _convert_job(job: SQLJob) -> UWSJob:
    """Convert the SQL representation of a job to its dataclass.

    The internal representation of a job uses a dataclass that is kept
    intentionally separate from the database schema so that the conversion can
    be done explicitly and the rest of the code kept separate from SQLAlchemy
    database models. This internal helper function converts from the database
    representation to the internal representation.
    """
    error = None
    if job.error_code and job.error_type and job.error_message:
        error = UWSJobError(
            error_type=job.error_type,
            error_code=job.error_code,
            message=job.error_message,
            detail=job.error_detail,
        )
    return UWSJob(
        job_id=str(job.id),
        message_id=job.message_id,
        owner=job.owner,
        phase=job.phase,
        run_id=job.run_id,
        creation_time=datetime_from_db(job.creation_time),
        start_time=datetime_from_db(job.start_time),
        end_time=datetime_from_db(job.end_time),
        destruction_time=datetime_from_db(job.destruction_time),
        execution_duration=timedelta(seconds=job.execution_duration),
        quote=job.quote,
        parameters=[
            UWSJobParameter(parameter_id=p.parameter, value=p.value)
            for p in sorted(job.parameters, key=lambda p: p.id)
        ],
        results=[
            UWSJobResult(
                result_id=r.result_id,
                url=r.url,
                size=r.size,
                mime_type=r.mime_type,
            )
            for r in sorted(job.results, key=lambda r: r.sequence)
        ],
        error=error,
    )


class JobStore:
    """Stores and manipulates jobs in the database.

    The canonical representation of any UWS job is in the database. This class
    provides methods to create, update, and delete UWS job records and their
    associated results and errors.

    Parameters
    ----------
    session
        The underlying database session.
    """

    def __init__(self, session: async_scoped_session) -> None:
        self._session = session

    async def add(
        self,
        *,
        owner: str,
        run_id: str | None = None,
        params: list[UWSJobParameter],
        execution_duration: timedelta,
        lifetime: timedelta,
    ) -> UWSJob:
        """Create a record of a new job.

        The job will be created in pending status.

        Parameters
        ----------
        owner
            The username of the owner of the job.
        run_id
            A client-supplied opaque identifier to record with the job.
        params
            The input parameters to the job.
        execution_duration
            The maximum length of time for which a job is allowed to run in
            seconds.
        lifetime
            The maximum lifetime of the job and its results, in seconds.
            After this time, any record of the job will be deleted.

        Returns
        -------
        safir.uws._models.Job
            The internal representation of the newly-created job.
        """
        now = current_datetime()
        destruction_time = now + lifetime
        sql_params = [
            SQLJobParameter(parameter=p.parameter_id, value=p.value)
            for p in params
        ]
        job = SQLJob(
            owner=owner,
            phase=ExecutionPhase.PENDING,
            run_id=run_id,
            creation_time=datetime_to_db(now),
            destruction_time=datetime_to_db(destruction_time),
            execution_duration=int(execution_duration.total_seconds()),
            parameters=sql_params,
            results=[],
        )
        async with self._session.begin():
            self._session.add_all([job, *sql_params])
            await self._session.flush()
            return _convert_job(job)

    async def availability(self) -> Availability:
        """Check that the database is up."""
        try:
            async with self._session.begin():
                await self._session.execute(select(SQLJob.id).limit(1))
            return Availability(available=True)
        except OperationalError:
            note = "cannot query UWS job database"
            return Availability(available=False, notes=[note])
        except Exception as e:
            note = f"{type(e).__name__}: {e!s}"
            return Availability(available=False, notes=[note])

    async def delete(self, job_id: str) -> None:
        """Delete a job by ID."""
        stmt = delete(SQLJob).where(SQLJob.id == int(job_id))
        async with self._session.begin():
            await self._session.execute(stmt)

    async def get(self, job_id: str) -> UWSJob:
        """Retrieve a job by ID."""
        async with self._session.begin():
            job = await self._get_job(job_id)
            return _convert_job(job)

    async def list_expired(self) -> list[UWSJobDescription]:
        """Delete all jobs that have passed their destruction time."""
        now = datetime_to_db(current_datetime())
        stmt = select(
            SQLJob.id,
            SQLJob.message_id,
            SQLJob.owner,
            SQLJob.phase,
            SQLJob.run_id,
            SQLJob.creation_time,
        ).where(SQLJob.destruction_time <= now)
        async with self._session.begin():
            jobs = await self._session.execute(stmt)
            return [
                UWSJobDescription(
                    job_id=str(j.id),
                    message_id=j.message_id,
                    owner=j.owner,
                    phase=j.phase,
                    run_id=j.run_id,
                    creation_time=datetime_from_db(j.creation_time),
                )
                for j in jobs.all()
            ]

    async def list_jobs(
        self,
        user: str,
        *,
        phases: list[ExecutionPhase] | None = None,
        after: datetime | None = None,
        count: int | None = None,
    ) -> list[UWSJobDescription]:
        """List the jobs for a particular user.

        Parameters
        ----------
        user
            Name of the user whose jobs to load.
        phases
            Limit the result to jobs in this list of possible execution
            phases.
        after
            Limit the result to jobs created after the given datetime in UTC.
        count
            Limit the results to the most recent count jobs.

        Returns
        -------
        list of safir.uws._models.JobDescription
            List of job descriptions matching the search criteria.
        """
        stmt = select(
            SQLJob.id,
            SQLJob.message_id,
            SQLJob.owner,
            SQLJob.phase,
            SQLJob.run_id,
            SQLJob.creation_time,
        ).where(SQLJob.owner == user)
        if phases:
            stmt = stmt.where(SQLJob.phase.in_(phases))
        if after:
            stmt = stmt.where(SQLJob.creation_time > datetime_to_db(after))
        stmt = stmt.order_by(SQLJob.creation_time.desc())
        if count:
            stmt = stmt.limit(count)
        async with self._session.begin():
            jobs = await self._session.execute(stmt)
            return [
                UWSJobDescription(
                    job_id=str(j.id),
                    message_id=j.message_id,
                    owner=j.owner,
                    phase=j.phase,
                    run_id=j.run_id,
                    creation_time=datetime_from_db(j.creation_time),
                )
                for j in jobs.all()
            ]

    @retry_async_transaction
    async def mark_aborted(self, job_id: str) -> None:
        """Mark a job as aborted.

        Parameters
        ----------
        job_id
            Identifier of the job.
        """
        async with self._session.begin():
            job = await self._get_job(job_id)
            job.phase = ExecutionPhase.ABORTED
            if job.start_time:
                job.end_time = datetime_to_db(current_datetime())

    @retry_async_transaction
    async def mark_completed(self, job_id: str, job_result: JobResult) -> None:
        """Mark a job as completed.

        Parameters
        ----------
        job_id
            Identifier of the job.
        job_result
            Result of the job.
        """
        end_time = job_result.finish_time.replace(microsecond=0)
        results = job_result.result
        if isinstance(results, Exception):
            await self.mark_failed(job_id, results, end_time=end_time)
            return

        async with self._session.begin():
            job = await self._get_job(job_id)
            job.end_time = datetime_to_db(end_time)
            if job.phase != ExecutionPhase.ABORTED:
                job.phase = ExecutionPhase.COMPLETED
                for sequence, result in enumerate(results, start=1):
                    sql_result = SQLJobResult(
                        job_id=job.id,
                        result_id=result.result_id,
                        sequence=sequence,
                        url=result.url,
                        size=result.size,
                        mime_type=result.mime_type,
                    )
                    self._session.add(sql_result)

    @retry_async_transaction
    async def mark_failed(
        self, job_id: str, exc: Exception, *, end_time: datetime | None = None
    ) -> None:
        """Mark a job as failed with an error.

        Parameters
        ----------
        job_id
            Identifier of the job.
        exc
            Exception of failed job.
        end_time
            When the job failed, if known.
        """
        if isinstance(exc, TaskError):
            error = exc.to_job_error()
        else:
            error = UWSJobError(
                error_type=ErrorType.FATAL,
                error_code=ErrorCode.ERROR,
                message="Unknown error executing task",
                detail=f"{type(exc).__name__}: {exc!s}",
            )
        async with self._session.begin():
            job = await self._get_job(job_id)
            job.end_time = datetime_to_db(end_time or current_datetime())
            if job.phase != ExecutionPhase.ABORTED:
                job.phase = ExecutionPhase.ERROR
                job.error_type = error.error_type
                job.error_code = error.error_code
                job.error_message = error.message
                job.error_detail = error.detail

    @retry_async_transaction
    async def mark_executing(self, job_id: str, start_time: datetime) -> None:
        """Mark a job as executing.

        Parameters
        ----------
        job_id
            Identifier of the job.
        start_time
            Time at which the job started executing.
        """
        start_time = start_time.replace(microsecond=0)
        async with self._session.begin():
            job = await self._get_job(job_id)
            if job.phase in (ExecutionPhase.PENDING, ExecutionPhase.QUEUED):
                job.phase = ExecutionPhase.EXECUTING
            job.start_time = datetime_to_db(start_time)

    @retry_async_transaction
    async def mark_queued(self, job_id: str, metadata: JobMetadata) -> None:
        """Mark a job as queued for processing.

        This is called by the web frontend after queuing the work. However,
        the worker may have gotten there first and have already updated the
        phase to executing, in which case we should not set it back to queued.

        Parameters
        ----------
        job_id
            Identifier of the job.
        metadata
            Metadata about the underlying arq job.
        """
        async with self._session.begin():
            job = await self._get_job(job_id)
            job.message_id = metadata.id
            if job.phase in (ExecutionPhase.PENDING, ExecutionPhase.HELD):
                job.phase = ExecutionPhase.QUEUED

    async def update_destruction(
        self, job_id: str, destruction: datetime
    ) -> None:
        """Update the destruction time of a job.

        Parameters
        ----------
        job_id
            Identifier of the job.
        destruction
            New destruction time.
        """
        destruction = destruction.replace(microsecond=0)
        async with self._session.begin():
            job = await self._get_job(job_id)
            job.destruction_time = datetime_to_db(destruction)

    async def update_execution_duration(
        self, job_id: str, execution_duration: timedelta
    ) -> None:
        """Update the destruction time of a job.

        Parameters
        ----------
        job_id
            Identifier of the job.
        execution_duration
            New execution duration.
        """
        async with self._session.begin():
            job = await self._get_job(job_id)
            job.execution_duration = int(execution_duration.total_seconds())

    async def _get_job(self, job_id: str) -> SQLJob:
        """Retrieve a job from the database by job ID."""
        stmt = select(SQLJob).where(SQLJob.id == int(job_id))
        job = (await self._session.execute(stmt)).scalar_one_or_none()
        if not job:
            raise UnknownJobError(job_id)
        return job



"""Test for long polling when retrieving jobs."""

from __future__ import annotations

import asyncio
from datetime import timedelta

import pytest
from httpx import AsyncClient

from safir.datetime import current_datetime, isodatetime
from safir.testing.uws import MockUWSJobRunner
from safir.uws import UWSJobParameter
from safir.uws._dependencies import UWSFactory
from safir.uws._models import UWSJobResult

PENDING_JOB = """
<uws:job
    version="1.1"
    xsi:schemaLocation="http://www.ivoa.net/xml/UWS/v1.0 UWS.xsd"
    xmlns:xml="http://www.w3.org/XML/1998/namespace"
    xmlns:uws="http://www.ivoa.net/xml/UWS/v1.0"
    xmlns:xlink="http://www.w3.org/1999/xlink"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
  <uws:jobId>1</uws:jobId>
  <uws:ownerId>user</uws:ownerId>
  <uws:phase>{}</uws:phase>
  <uws:creationTime>{}</uws:creationTime>
  <uws:executionDuration>600</uws:executionDuration>
  <uws:destruction>{}</uws:destruction>
  <uws:parameters>
    <uws:parameter id="name">Naomi</uws:parameter>
  </uws:parameters>
</uws:job>
"""

EXECUTING_JOB = """
<uws:job
    version="1.1"
    xsi:schemaLocation="http://www.ivoa.net/xml/UWS/v1.0 UWS.xsd"
    xmlns:xml="http://www.w3.org/XML/1998/namespace"
    xmlns:uws="http://www.ivoa.net/xml/UWS/v1.0"
    xmlns:xlink="http://www.w3.org/1999/xlink"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
  <uws:jobId>1</uws:jobId>
  <uws:ownerId>user</uws:ownerId>
  <uws:phase>EXECUTING</uws:phase>
  <uws:creationTime>{}</uws:creationTime>
  <uws:startTime>{}</uws:startTime>
  <uws:executionDuration>600</uws:executionDuration>
  <uws:destruction>{}</uws:destruction>
  <uws:parameters>
    <uws:parameter id="name">Naomi</uws:parameter>
  </uws:parameters>
</uws:job>
"""

FINISHED_JOB = """
<uws:job
    version="1.1"
    xsi:schemaLocation="http://www.ivoa.net/xml/UWS/v1.0 UWS.xsd"
    xmlns:xml="http://www.w3.org/XML/1998/namespace"
    xmlns:uws="http://www.ivoa.net/xml/UWS/v1.0"
    xmlns:xlink="http://www.w3.org/1999/xlink"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
  <uws:jobId>1</uws:jobId>
  <uws:ownerId>user</uws:ownerId>
  <uws:phase>COMPLETED</uws:phase>
  <uws:creationTime>{}</uws:creationTime>
  <uws:startTime>{}</uws:startTime>
  <uws:endTime>{}</uws:endTime>
  <uws:executionDuration>600</uws:executionDuration>
  <uws:destruction>{}</uws:destruction>
  <uws:parameters>
    <uws:parameter id="name">Naomi</uws:parameter>
  </uws:parameters>
  <uws:results>
    <uws:result id="cutout" xlink:href="https://example.com/some/path"\
 mime-type="application/fits"/>
  </uws:results>
</uws:job>
"""


@pytest.mark.asyncio
async def test_poll(
    client: AsyncClient, runner: MockUWSJobRunner, uws_factory: UWSFactory
) -> None:
    job_service = uws_factory.create_job_service()
    job = await job_service.create(
        "user",
        params=[UWSJobParameter(parameter_id="name", value="Naomi")],
    )

    # Poll for changes for one second. Nothing will happen since nothing is
    # changing the mock arq queue.
    now = current_datetime()
    r = await client.get(
        "/test/jobs/1",
        headers={"X-Auth-Request-User": "user"},
        params={"WAIT": "1"},
    )
    assert (current_datetime() - now).total_seconds() >= 1
    assert r.status_code == 200
    assert r.text == PENDING_JOB.strip().format(
        "PENDING",
        isodatetime(job.creation_time),
        isodatetime(job.creation_time + timedelta(seconds=24 * 60 * 60)),
    )

    # Start the job and worker.
    r = await client.post(
        "/test/jobs/1/phase",
        headers={"X-Auth-Request-User": "user"},
        data={"PHASE": "RUN"},
        follow_redirects=True,
    )
    assert r.status_code == 200
    assert r.url == "https://example.com/test/jobs/1"
    assert r.text == PENDING_JOB.strip().format(
        "QUEUED",
        isodatetime(job.creation_time),
        isodatetime(job.creation_time + timedelta(seconds=24 * 60 * 60)),
    )

    # Poll for a change from queued, which we should see after half a second.
    now = current_datetime()
    job, r = await asyncio.gather(
        runner.mark_in_progress("user", "1", delay=0.5),
        client.get(
            "/test/jobs/1",
            headers={"X-Auth-Request-User": "user"},
            params={"WAIT": "2", "phase": "QUEUED"},
        ),
    )
    assert r.status_code == 200
    assert job.start_time
    assert r.text == EXECUTING_JOB.strip().format(
        isodatetime(job.creation_time),
        isodatetime(job.start_time),
        isodatetime(job.creation_time + timedelta(seconds=24 * 60 * 60)),
    )

    # Now, wait again, in parallel with the job finishing. We should get a
    # reply after a second and a half when the job finishes.
    results = [
        UWSJobResult(
            result_id="cutout",
            url="s3://some-bucket/some/path",
            mime_type="application/fits",
        )
    ]
    job, r = await asyncio.gather(
        runner.mark_complete("user", "1", results, delay=1.5),
        client.get(
            "/test/jobs/1",
            headers={"X-Auth-Request-User": "user"},
            params={"WAIT": "2", "phase": "EXECUTING"},
        ),
    )
    assert r.status_code == 200
    assert job.start_time
    assert job.end_time
    assert r.text == FINISHED_JOB.strip().format(
        isodatetime(job.creation_time),
        isodatetime(job.start_time),
        isodatetime(job.end_time),
        isodatetime(job.creation_time + timedelta(seconds=24 * 60 * 60)),
    )
    assert (current_datetime() - now).total_seconds() >= 2



"""Representation of a UWS application."""

from __future__ import annotations

from dataclasses import asdict
from pathlib import Path
from typing import Any

from arq import cron
from fastapi import APIRouter, FastAPI, Request
from fastapi.exceptions import RequestValidationError
from fastapi.responses import PlainTextResponse
from structlog.stdlib import BoundLogger

from safir.arq import ArqQueue, WorkerSettings
from safir.arq.uws import UWS_QUEUE_NAME
from safir.database import (
    create_database_engine,
    initialize_database,
    is_database_current,
    stamp_database_async,
)
from safir.middleware.ivoa import CaseInsensitiveQueryMiddleware

from ._config import UWSConfig
from ._constants import UWS_DATABASE_TIMEOUT, UWS_EXPIRE_JOBS_SCHEDULE
from ._dependencies import uws_dependency
from ._exceptions import DatabaseSchemaError, UWSError
from ._handlers import (
    install_async_post_handler,
    install_availability_handler,
    install_sync_get_handler,
    install_sync_post_handler,
    uws_router,
)
from ._schema import UWSSchemaBase
from ._workers import (
    close_uws_worker_context,
    create_uws_worker_context,
    uws_expire_jobs,
    uws_job_completed,
    uws_job_started,
)

__all__ = ["UWSApplication"]


async def _uws_error_handler(
    request: Request, exc: UWSError
) -> PlainTextResponse:
    response = f"{exc.error_code.value}: {exc!s}\n"
    if exc.detail:
        response += "\n{exc.detail}"
    return PlainTextResponse(response, status_code=exc.status_code)


async def _usage_handler(
    request: Request, exc: RequestValidationError
) -> PlainTextResponse:
    return PlainTextResponse(f"UsageError\n\n{exc!s}", status_code=422)


class UWSApplication:
    """Glue between a FastAPI application and the UWS implementation.

    An instance of this class should be created during construction of the
    service that will use the UWS layer. It provides methods to initialize the
    UWS database, build route handlers, install error handlers, and build the
    UWS database worker. Construction of the backend worker that does the work
    of the service is handled separately so that it can have minimal
    dependencies.

    Parameters
    ----------
    config
        UWS configuration.
    """

    def __init__(self, config: UWSConfig) -> None:
        self._config = config

    def build_worker(
        self,
        logger: BoundLogger,
        *,
        check_schema: bool = False,
        alembic_config_path: Path = Path("alembic.ini"),
    ) -> WorkerSettings:
        """Construct an arq worker configuration for the UWS worker.

        All UWS job status and results must be stored in the underlying
        database, since the API serves job information from there. To minimize
        dependencies for the worker, which may (for example) pin its own
        version of SQLAlchemy that may not be compatible with that used by the
        application, the actual worker is not responsible for storing the
        results in SQL. Instead, it returns results via arq, which temporarily
        puts them in Redis then uses ``on_job_start`` and ``after_job_end`` to
        notify a different queue. Those results are recovered and stored in
        the database by separate a separate arq worker.

        This function returns a class suitable for assigning to a module
        variable and referencing as the argument to the :command:`arq`
        command-line tool to start the worker.

        Parameters
        ----------
        logger
            Logger to use for messages.
        check_schema
            Whether to check the database schema version with Alembic on
            startup.
        alembic_config_path
            When checking the schema, use this path to the Alembic
            configuration.
        """

        async def startup(ctx: dict[Any, Any]) -> None:
            ctx.update(
                await create_uws_worker_context(
                    self._config,
                    logger,
                    check_schema=check_schema,
                    alembic_config_path=alembic_config_path,
                )
            )

        async def shutdown(ctx: dict[Any, Any]) -> None:
            await close_uws_worker_context(ctx)

        # Running 10 jobs simultaneously is the arq default as of arq 0.26.0
        # and seems reasonable for database workers.
        return WorkerSettings(
            cron_jobs=[
                cron(
                    uws_expire_jobs,
                    unique=True,
                    timeout=UWS_DATABASE_TIMEOUT,
                    **asdict(UWS_EXPIRE_JOBS_SCHEDULE),
                )
            ],
            functions=[uws_job_started, uws_job_completed],
            redis_settings=self._config.arq_redis_settings,
            job_completion_wait=UWS_DATABASE_TIMEOUT,
            job_timeout=UWS_DATABASE_TIMEOUT,
            max_jobs=10,
            queue_name=UWS_QUEUE_NAME,
            on_startup=startup,
            on_shutdown=shutdown,
        )

    async def initialize_fastapi(
        self,
        logger: BoundLogger | None = None,
        *,
        check_schema: bool = False,
        alembic_config_path: Path = Path("alembic.ini"),
    ) -> None:
        """Initialize the UWS subsystem for FastAPI applications.

        This must be called before any UWS routes are accessed, normally from
        the lifespan function of the FastAPI application.

        Parameters
        ----------
        logger
            Logger to use to report any problems.
        check_schema
            If `True`, check whether the database schema for the UWS database
            is up to date using Alembic.
        alembic_config_path
            When checking the schema, use this path to the Alembic
            configuration.

        Raises
        ------
        DatabaseSchemaError
            Raised if the UWS database schema is out of date.
        """
        if check_schema:
            if not await self.is_schema_current(logger, alembic_config_path):
                raise DatabaseSchemaError("UWS database schema out of date")
        await uws_dependency.initialize(self._config)

    async def initialize_uws_database(
        self,
        logger: BoundLogger,
        *,
        reset: bool = False,
        use_alembic: bool = False,
        alembic_config_path: Path = Path("alembic.ini"),
    ) -> None:
        """Initialize the UWS database.

        Parameters
        ----------
        logger
            Logger to use.
        reset
            If `True`, also delete all data in the database.
        use_alembic
            Whether to stamp the UWS database with Alembic.
        alembic_config_path
            When stamping the database, use this path to the Alembic
            configuration.
        """
        engine = create_database_engine(
            self._config.database_url, self._config.database_password
        )
        await initialize_database(
            engine, logger, schema=UWSSchemaBase.metadata, reset=reset
        )
        if use_alembic:
            await stamp_database_async(engine, alembic_config_path)
        await engine.dispose()

    def install_error_handlers(self, app: FastAPI) -> None:
        """Install error handlers that follow DALI and UWS conventions.

        This method must be called during application setup for any FastAPI
        app using the UWS layer for correct error message handling. This will
        change the error response for all parameter validation errors from
        FastAPI.

        Currently these error handlers return ``text/plain`` errors. VOTable
        errors may be a better choice, but revision 1.0 of the SODA standard
        only allows ``text/plain`` errors for sync routes.
        """
        app.exception_handler(UWSError)(_uws_error_handler)
        app.exception_handler(RequestValidationError)(_usage_handler)

    def install_handlers(self, router: APIRouter) -> None:
        """Install the route handlers for the service.

        This method will always install a POST handler at the root of the
        router that creates an async job, and handlers under ``/jobs`` that
        implement the UWS protocol for managing those jobs. If
        ``sync_post_route`` is set in the `~safir.uws.UWSConfig` that this
        application was configured with, a POST handler for ``/sync`` to
        create a sync job will be added. If ``sync_get_route`` is set, a GET
        handler for ``/sync`` to create a sync job will be added.
        """
        router.include_router(uws_router, prefix="/jobs")
        install_availability_handler(router)
        if route := self._config.sync_get_route:
            install_sync_get_handler(router, route)
        if route := self._config.sync_post_route:
            install_sync_post_handler(router, route)

        # This handler must be installed directly on the provided router. Do
        # not install it on the UWS router before include_router. The process
        # of copying handlers done by include_router loses the dependency
        # information from the async post handler and causes it to not see any
        # of its job parameters.
        #
        # This is probably because the dependency is a dynamic function not
        # known statically, which may confuse the handler copying code in
        # FastAPI. This problem was last verified in FastAPI 0.111.0.
        install_async_post_handler(router, self._config.async_post_route)

    def install_middleware(self, app: FastAPI) -> None:
        """Install FastAPI middleware needed by UWS.

        UWS unfortunately requires that the key portion of query parameters be
        case-insensitive, so UWS FastAPI applications need to add custom
        middleware to lowercase query parameter keys. This method does that.

        Parameters
        ----------
        app
            FastAPI app.
        """
        app.add_middleware(CaseInsensitiveQueryMiddleware)

    async def is_schema_current(
        self,
        logger: BoundLogger | None = None,
        config_path: Path = Path("alembic.ini"),
    ) -> bool:
        """Check that the database schema is current using Alembic.

        Parameters
        ----------
        logger
            Logger to use to report any problems.
        config_path
            Path to the Alembic configuration.
        """
        engine = create_database_engine(
            self._config.database_url, self._config.database_password
        )
        try:
            return await is_database_current(engine, logger, config_path)
        finally:
            await engine.dispose()

    def override_arq_queue(self, arq_queue: ArqQueue) -> None:
        """Change the arq used by the FastAPI route handlers.

        This method is probably only useful for the test suite.

        Parameters
        ----------
        arq_queue
            New arq queue.
        """
        uws_dependency.override_arq_queue(arq_queue)

    async def shutdown_fastapi(self) -> None:
        """Shut down the UWS subsystem for FastAPI applications.

        This should be called during application shutdown, normally from the
        lifespan function of the FastAPI application.
        """
        await uws_dependency.aclose()



"""Support library for writing UWS-enabled services."""

from ._app import UWSApplication
from ._config import ParametersModel, UWSAppSettings, UWSConfig, UWSRoute
from ._dependencies import uws_post_params_dependency
from ._exceptions import (
    DatabaseSchemaError,
    MultiValuedParameterError,
    ParameterError,
    ParameterParseError,
    UsageError,
    UWSError,
)
from ._models import (
    ErrorCode,
    UWSJob,
    UWSJobError,
    UWSJobParameter,
    UWSJobResult,
)
from ._schema import UWSSchemaBase

__all__ = [
    "DatabaseSchemaError",
    "ErrorCode",
    "MultiValuedParameterError",
    "ParameterError",
    "ParameterParseError",
    "ParametersModel",
    "UWSApplication",
    "UWSAppSettings",
    "UWSConfig",
    "UWSError",
    "UWSJob",
    "UWSJobError",
    "UWSJobParameter",
    "UWSJobResult",
    "UWSRoute",
    "UWSSchemaBase",
    "UsageError",
    "uws_post_params_dependency",
]



"""Tests for the database utility functions."""

from __future__ import annotations

import asyncio
import os
import subprocess
from datetime import UTC, datetime, timedelta, timezone
from pathlib import Path
from typing import Any
from urllib.parse import unquote, urlparse

import pytest
import structlog
from pydantic import BaseModel, SecretStr
from pydantic_core import Url
from sqlalchemy import Column, MetaData, String, Table
from sqlalchemy.exc import OperationalError, ProgrammingError
from sqlalchemy.future import select

from safir.database import (
    create_async_session,
    create_database_engine,
    datetime_from_db,
    datetime_to_db,
    initialize_database,
    is_database_current,
    retry_async_transaction,
    stamp_database,
    stamp_database_async,
    unstamp_database,
)
from safir.database._connection import build_database_url

from .support.alembic import BaseV1, BaseV2, UserV1, UserV2, config


@pytest.mark.asyncio
async def test_database_init(
    database_url: str, database_password: str
) -> None:
    logger = structlog.get_logger(__name__)
    engine = create_database_engine(database_url, database_password)
    await initialize_database(
        engine, logger, schema=BaseV2.metadata, reset=True
    )
    session = await create_async_session(engine, logger)
    async with session.begin():
        session.add(UserV2(username="someuser"))
    await session.remove()

    # Reinitializing the database without reset should preserve the row.
    await initialize_database(engine, logger, schema=BaseV2.metadata)
    session = await create_async_session(engine, logger)
    async with session.begin():
        result = await session.scalars(select(UserV2.username))
        assert result.all() == ["someuser"]
    await session.remove()

    # Reinitializing the database with reset should delete the data. Try
    # passing in the password as a SecretStr.
    password = SecretStr(database_password)
    engine = create_database_engine(database_url, password)
    await initialize_database(
        engine, logger, schema=BaseV2.metadata, reset=True
    )
    session = await create_async_session(engine, logger)
    async with session.begin():
        result = await session.scalars(select(UserV2.username))
        assert result.all() == []
    await session.remove()
    await engine.dispose()


def test_build_database_url(database_url: str) -> None:
    url = build_database_url("postgresql://foo@127.0.0.1/foo", None)
    assert url == "postgresql+asyncpg://foo@127.0.0.1/foo"

    url = build_database_url("postgresql://foo@127.0.0.1:5432/foo", None)
    assert url == "postgresql+asyncpg://foo@127.0.0.1:5432/foo"

    url = build_database_url("postgresql://foo@127.0.0.1/foo", "otherpass")
    assert url == "postgresql+asyncpg://foo:otherpass@127.0.0.1/foo"

    url = build_database_url(
        "postgresql://foo@127.0.0.1:5433/foo", "otherpass"
    )
    assert url == "postgresql+asyncpg://foo:otherpass@127.0.0.1:5433/foo"

    pydantic_url = Url.build(
        scheme="postgresql", username="user", host="localhost", path="foo"
    )
    url = build_database_url(pydantic_url, "password")
    assert url == "postgresql+asyncpg://user:password@localhost/foo"

    # Test that the username and password are quoted properly.
    url = build_database_url(
        "postgresql://foo%40e.com@127.0.0.1:4444/foo",
        "pass@word/with stuff",
    )
    assert url == (
        "postgresql+asyncpg://foo%40e.com:pass%40word%2Fwith%20stuff"
        "@127.0.0.1:4444/foo"
    )
    parsed_url = urlparse(url)
    assert parsed_url.username
    assert parsed_url.password

    # urlparse does not undo quoting in the components of netloc.
    assert unquote(parsed_url.username) == "foo@e.com"
    assert unquote(parsed_url.password) == "pass@word/with stuff"
    assert parsed_url.hostname == "127.0.0.1"
    assert parsed_url.port == 4444


@pytest.mark.asyncio
async def test_create_async_session(
    database_url: str, database_password: str
) -> None:
    logger = structlog.get_logger(__name__)
    engine = create_database_engine(database_url, database_password)
    await initialize_database(
        engine, logger, schema=BaseV2.metadata, reset=True
    )

    session = await create_async_session(
        engine, logger, statement=select(UserV2)
    )
    async with session.begin():
        session.add(UserV2(username="foo"))
    await session.remove()

    # Use a query against a non-existent table as the liveness check and
    # ensure that fails.
    metadata = MetaData()
    bad_table = Table("bad", metadata, Column("name", String(64)))
    with pytest.raises(ProgrammingError):
        session = await create_async_session(
            engine, logger, statement=select(bad_table)
        )
    await engine.dispose()


def test_datetime() -> None:
    tz_aware = datetime.now(tz=UTC)
    tz_naive = tz_aware.replace(tzinfo=None)

    assert datetime_to_db(tz_aware) == tz_naive
    assert datetime_from_db(tz_naive) == tz_aware
    assert datetime_from_db(tz_aware) == tz_aware

    assert datetime_to_db(None) is None
    assert datetime_from_db(None) is None

    with pytest.raises(ValueError, match=r"datetime .* not in UTC"):
        datetime_to_db(tz_naive)

    tz_local = datetime.now(tz=timezone(timedelta(hours=1)))
    with pytest.raises(ValueError, match=r"datetime .* not in UTC"):
        datetime_to_db(tz_local)
    with pytest.raises(ValueError, match=r"datetime .* not in UTC"):
        datetime_from_db(tz_local)

    # Pydantic's JSON decoder uses a TzInfo data structure instead of
    # datetime.timezone.utc. Make sure that's still recognized as UTC.
    class Test(BaseModel):
        time: datetime

    json_model = Test(time=tz_aware).model_dump_json()
    model = Test.model_validate_json(json_model)
    assert datetime_to_db(model.time) == tz_naive


@pytest.mark.asyncio
async def test_retry_async_transaction(
    database_url: str, database_password: str
) -> None:
    logger = structlog.get_logger(__name__)
    engine = create_database_engine(database_url, database_password)
    await initialize_database(
        engine, logger, schema=BaseV2.metadata, reset=True
    )
    session = await create_async_session(engine, logger)
    async with session.begin():
        session.add(UserV2(username="someuser"))
    tries = 0

    @retry_async_transaction
    async def insert(attempts: int) -> None:
        nonlocal tries
        tries += 1
        async with session.begin():
            if tries <= attempts:
                raise OperationalError(None, None, ValueError("foo"))
            session.add(UserV2(username="newuser"))

    await insert(2)

    tries = 0
    with pytest.raises(OperationalError):
        await insert(3)

    @retry_async_transaction(max_tries=1)
    async def insert_capped(attempts: int) -> None:
        nonlocal tries
        tries += 1
        async with session.begin():
            if tries <= attempts:
                raise OperationalError(None, None, ValueError("foo"))
            session.add(UserV2(username="newuser"))

    tries = 0
    with pytest.raises(OperationalError):
        await insert_capped(1)

    await session.remove()
    await engine.dispose()


def test_alembic(
    database_url: str,
    database_password: str,
    monkeypatch: pytest.MonkeyPatch,
    event_loop: asyncio.AbstractEventLoop,
) -> None:
    config.database_url = Url(database_url)
    config.database_password = SecretStr(database_password)
    config_path = (
        Path(__file__).parent / "data" / "database" / "v1" / "alembic.ini"
    )
    logger = structlog.get_logger(__name__)
    engine = create_database_engine(database_url, database_password)

    async def init(schema: MetaData) -> None:
        await initialize_database(engine, logger, schema=schema, reset=True)

    async def check(config_path: Path) -> bool:
        return await is_database_current(engine, logger, config_path)

    async def store(record: Any) -> None:
        session = await create_async_session(engine, logger)
        async with session.begin():
            session.add(record)
        await session.remove()

    async def list_v2() -> list[str]:
        session = await create_async_session(engine, logger)
        async with session.begin():
            result = await session.scalars(select(UserV2.username))
        await session.remove()
        return list(result.all())

    # Initialize the database with the V1 schema.
    event_loop.run_until_complete(init(BaseV1.metadata))

    # Before stamping, the database should show as not up to date.
    assert not event_loop.run_until_complete(check(config_path))

    # Stamp the database.
    stamp_database(config_path)

    # Check that the database is up to date.
    assert event_loop.run_until_complete(check(config_path))

    # Store a record in the database.
    event_loop.run_until_complete(store(UserV1(username="foo", name="Foo")))

    # Check that the database is not current with the V2 schema.
    config_path = config_path.parent.parent / "v2" / "alembic.ini"
    assert not event_loop.run_until_complete(check(config_path))

    # Upgrade the database to V2. It should now show as current with the V2
    # schema.
    subprocess.run(
        ["alembic", "upgrade", "head"],
        check=True,
        cwd=str(config_path.parent),
        env={
            **os.environ,
            "TEST_DATABASE_URL": database_url,
            "TEST_DATABASE_PASSWORD": database_password,
            "PYTHONPATH": str(Path(__file__).parent.parent),
        },
    )
    assert event_loop.run_until_complete(check(config_path))

    # The data should still be there.
    assert event_loop.run_until_complete(list_v2()) == ["foo"]

    # Unstamping the database should cause it to no longer be current, but we
    # should be able to them stamp it async and then it should be current
    # again.
    event_loop.run_until_complete(unstamp_database(engine))
    assert not event_loop.run_until_complete(check(config_path))
    event_loop.run_until_complete(stamp_database_async(engine, config_path))
    assert event_loop.run_until_complete(check(config_path))



"""Constants for the UWS work queue."""

from __future__ import annotations

from datetime import timedelta

from arq.cron import Options

__all__ = [
    "JOB_STOP_TIMEOUT",
    "JOB_RESULT_TIMEOUT",
    "UWS_DATABASE_TIMEOUT",
    "UWS_EXPIRE_JOBS_SCHEDULE",
]

JOB_RESULT_TIMEOUT = timedelta(seconds=5)
"""How long to poll arq for job results before giving up."""

JOB_STOP_TIMEOUT = timedelta(seconds=30)
"""How long to wait for a job to stop before giving up."""

UWS_DATABASE_TIMEOUT = timedelta(seconds=30)
"""Timeout on workers that update the UWS database.

This should match the default Kubernetes grace period for a pod to shut down.
"""

UWS_EXPIRE_JOBS_SCHEDULE = Options(
    month=None,
    day=None,
    weekday=None,
    hour=None,
    minute=5,
    second=0,
    microsecond=0,
)
"""Schedule for job expiration cron job, as `arq.cron.cron` parameters."""



"""Utility functions for database management."""

from ._alembic import (
    AlembicConfigError,
    is_database_current,
    run_migrations_offline,
    run_migrations_online,
    stamp_database,
    stamp_database_async,
    unstamp_database,
)
from ._connection import create_async_session, create_database_engine
from ._datetime import datetime_from_db, datetime_to_db
from ._initialize import DatabaseInitializationError, initialize_database
from ._retry import RetryP, RetryT, retry_async_transaction

__all__ = [
    "AlembicConfigError",
    "DatabaseInitializationError",
    "create_async_session",
    "create_database_engine",
    "datetime_from_db",
    "datetime_to_db",
    "initialize_database",
    "is_database_current",
    "retry_async_transaction",
    "run_migrations_offline",
    "run_migrations_online",
    "stamp_database",
    "stamp_database_async",
    "unstamp_database",
    # Included only for documentation purposes.
    "RetryP",
    "RetryT",
]



"""FastAPI dependencies for the UWS service.

The UWS FastAPI support is initialized by the parent application via this
dependency's ``initialize`` method. It then returns a `UWSFactory` on request
to individual route handlers, which in turn can create other needed objects.
"""

from collections.abc import AsyncIterator
from typing import Annotated, Literal

from fastapi import Depends, Form, Query, Request
from sqlalchemy.ext.asyncio import AsyncEngine, async_scoped_session
from structlog.stdlib import BoundLogger

from safir.arq import ArqMode, ArqQueue, MockArqQueue, RedisArqQueue
from safir.database import create_async_session, create_database_engine
from safir.dependencies.logger import logger_dependency

from ._config import UWSConfig
from ._exceptions import ParameterError
from ._models import UWSJobParameter
from ._responses import UWSTemplates
from ._results import ResultStore
from ._service import JobService
from ._storage import JobStore

__all__ = [
    "UWSDependency",
    "UWSFactory",
    "create_phase_dependency",
    "runid_post_dependency",
    "uws_dependency",
    "uws_post_params_dependency",
]


class UWSFactory:
    """Build UWS components.

    Parameters
    ----------
    config
        UWS configuration.
    arq
        arq queue to use.
    session
        Database session.
    result_store
        Signed URL generator for results.
    logger
        Logger to use.

    Attributes
    ----------
    session
        Database session. This is exposed primarily for the test suite. It
        shouldn't be necessary for other code to use it directly.
    """

    def __init__(
        self,
        *,
        config: UWSConfig,
        arq: ArqQueue,
        session: async_scoped_session,
        result_store: ResultStore,
        logger: BoundLogger,
    ) -> None:
        self.session = session
        self._config = config
        self._arq = arq
        self._result_store = result_store
        self._logger = logger

    def create_result_store(self) -> ResultStore:
        """Return a wrapper around the result storage."""
        return self._result_store

    def create_job_service(self) -> JobService:
        """Create a new UWS job metadata service."""
        return JobService(
            config=self._config,
            arq_queue=self._arq,
            storage=self.create_job_store(),
            logger=self._logger,
        )

    def create_job_store(self) -> JobStore:
        """Create a new UWS job store."""
        return JobStore(self.session)

    def create_templates(self) -> UWSTemplates:
        """Create a new XML renderer for responses."""
        return UWSTemplates(self._result_store)


class UWSDependency:
    """Initializes UWS and provides a UWS factory as a dependency."""

    def __init__(self) -> None:
        self._arq: ArqQueue | None = None
        self._config: UWSConfig
        self._engine: AsyncEngine
        self._session: async_scoped_session
        self._result_store: ResultStore

    async def __call__(
        self, logger: Annotated[BoundLogger, Depends(logger_dependency)]
    ) -> AsyncIterator[UWSFactory]:
        if not self._arq:
            raise RuntimeError("UWSDependency not initialized")
        try:
            yield UWSFactory(
                config=self._config,
                arq=self._arq,
                session=self._session,
                result_store=self._result_store,
                logger=logger,
            )
        finally:
            # Following the recommendations in the SQLAlchemy documentation,
            # each session is scoped to a single web request. However, this
            # all uses the same async_scoped_session object, so should share
            # an underlying engine and connection pool.
            await self._session.remove()

    async def aclose(self) -> None:
        """Shut down the UWS subsystem."""
        await self._engine.dispose()

    async def initialize(self, config: UWSConfig) -> None:
        """Initialize the UWS subsystem.

        Parameters
        ----------
        config
            UWS configuration.
        """
        self._config = config
        self._result_store = ResultStore(config)
        if not self._arq:
            if config.arq_mode == ArqMode.production:
                settings = config.arq_redis_settings
                self._arq = await RedisArqQueue.initialize(settings)
            else:
                self._arq = MockArqQueue()
        self._engine = create_database_engine(
            config.database_url,
            config.database_password,
            isolation_level="REPEATABLE READ",
        )
        self._session = await create_async_session(self._engine)

    def override_arq_queue(self, arq_queue: ArqQueue) -> None:
        """Change the arq used in subsequent invocations.

        This method is probably only useful for the test suite.

        Parameters
        ----------
        arq_queue
            New arq queue.
        """
        self._arq = arq_queue


uws_dependency = UWSDependency()


async def uws_post_params_dependency(
    request: Request,
) -> list[UWSJobParameter]:
    """Parse POST parameters.

    UWS requires that all POST parameters be case-insensitive, which is not
    supported by FastAPI or Starlette. POST parameters therefore have to be
    parsed by this dependency and then extracted from the resulting
    `~safir.uws.UWSJobParameter` list (which unfortunately also means
    revalidating their types).

    The POST parameters can also be (and should be) listed independently as
    dependencies using the normal FastAPI syntax, in order to populate the
    OpenAPI schema, but unfortunately they all have to be listed as optional
    from FastAPI's perspective because they may be present using different
    capitalization.
    """
    if request.method != "POST":
        raise ValueError("uws_post_params_dependency used for non-POST route")
    parameters = []
    for key, value in (await request.form()).items():
        if not isinstance(value, str):
            raise TypeError("File upload not supported")
        parameters.append(
            UWSJobParameter(parameter_id=key.lower(), value=value)
        )
    return parameters


async def create_phase_dependency(
    *,
    get_phase: Annotated[
        Literal["RUN"] | None,
        Query(title="Immediately start job", alias="phase"),
    ] = None,
    post_phase: Annotated[
        Literal["RUN"] | None,
        Form(title="Immediately start job", alias="phase"),
    ] = None,
    params: Annotated[
        list[UWSJobParameter], Depends(uws_post_params_dependency)
    ],
) -> Literal["RUN"] | None:
    """Parse the optional phase parameter to an async job creation.

    Allow ``phase=RUN`` to be specified in either the query or the POST
    parameters, which says that the job should be immediately started.
    """
    for param in params:
        if param.parameter_id != "phase":
            continue
        if param.value != "RUN":
            raise ParameterError(f"Invalid phase {param.value}")
        return "RUN"
    return get_phase


async def runid_post_dependency(
    *,
    runid: Annotated[
        str | None,
        Form(
            title="Run ID for job",
            description=(
                "An opaque string that is returned in the job metadata and"
                " job listings. May be used by the client to associate jobs"
                " with specific larger operations."
            ),
        ),
    ] = None,
    params: Annotated[
        list[UWSJobParameter], Depends(uws_post_params_dependency)
    ],
) -> str | None:
    """Parse the run ID from POST parameters.

    This is annoyingly complex because DALI defines all parameters as
    case-insensitive, so we have to list the field as a dependency but then
    parse it out of the case-canonicalized job parameters.
    """
    for param in params:
        if param.parameter_id == "runid":
            runid = param.value
    return runid



"""Models for UWS services.

See https://www.ivoa.net/documents/UWS/20161024/REC-UWS-1.1-20161024.html for
most of these models. Descriptive language here is paraphrased from this
standard.
"""

from __future__ import annotations

from dataclasses import asdict, dataclass
from datetime import datetime, timedelta
from enum import StrEnum

from vo_models.uws import (
    ErrorSummary,
    JobSummary,
    Parameter,
    Parameters,
    ResultReference,
    ShortJobDescription,
)
from vo_models.uws.types import ErrorType, ExecutionPhase, UWSVersion

__all__ = [
    "ACTIVE_PHASES",
    "ErrorCode",
    "UWSJob",
    "UWSJobDescription",
    "UWSJobError",
    "UWSJobParameter",
    "UWSJobResult",
    "UWSJobResultSigned",
]


ACTIVE_PHASES = {
    ExecutionPhase.PENDING,
    ExecutionPhase.QUEUED,
    ExecutionPhase.EXECUTING,
}
"""Phases in which the job is active and can be waited on."""


class ErrorCode(StrEnum):
    """Possible error codes in ``text/plain`` responses.

    The following list of errors is taken from the SODA specification and
    therefore may not be appropriate for all DALI services.
    """

    AUTHENTICATION_ERROR = "AuthenticationError"
    AUTHORIZATION_ERROR = "AuthorizationError"
    MULTIVALUED_PARAM_NOT_SUPPORTED = "MultiValuedParamNotSupported"
    ERROR = "Error"
    SERVICE_UNAVAILABLE = "ServiceUnavailable"
    USAGE_ERROR = "UsageError"


@dataclass
class UWSJobError:
    """Failure information about a job."""

    error_type: ErrorType
    """Type of the error."""

    error_code: ErrorCode
    """The SODA error code of this error."""

    message: str
    """Brief error message.

    Note that the UWS specification allows a sequence of messages, but we only
    use a single message and thus a sequence of length one.
    """

    detail: str | None = None
    """Extended error message with additional detail."""

    def to_dict(self) -> dict[str, str | None]:
        """Convert to a dictionary, primarily for logging."""
        return asdict(self)

    def to_xml_model(self) -> ErrorSummary:
        """Convert to a Pydantic XML model."""
        return ErrorSummary(
            message=f"{self.error_code.value}: {self.message}",
            type=self.error_type,
            has_detail=self.detail is not None,
        )


@dataclass
class UWSJobResult:
    """A single result from the job."""

    result_id: str
    """Identifier for the result."""

    url: str
    """The URL for the result, which must point into a GCS bucket."""

    size: int | None = None
    """Size of the result in bytes."""

    mime_type: str | None = None
    """MIME type of the result."""

    def to_xml_model(self) -> ResultReference:
        """Convert to a Pydantic XML model."""
        return ResultReference(
            id=self.result_id, size=self.size, mime_type=self.mime_type
        )


@dataclass
class UWSJobResultSigned:
    """A single result from the job with a signed URL.

    A `UWSJobResult` is converted to a `UWSJobResultSigned` before generating
    the response via templating or returning the URL as a redirect.
    """

    result_id: str
    """Identifier for the result."""

    url: str
    """Signed URL to retrieve the result."""

    size: int | None = None
    """Size of the result in bytes."""

    mime_type: str | None = None
    """MIME type of the result."""

    def to_xml_model(self) -> ResultReference:
        """Convert to a Pydantic XML model."""
        return ResultReference(
            id=self.result_id,
            type=None,
            href=self.url,
            size=self.size,
            mime_type=self.mime_type,
        )


@dataclass
class UWSJobParameter:
    """An input parameter to the job."""

    parameter_id: str
    """Identifier of the parameter."""

    value: str
    """Value of the parameter."""

    def to_dict(self) -> dict[str, str | bool]:
        """Convert to a dictionary, primarily for logging."""
        return asdict(self)

    def to_xml_model(self) -> Parameter:
        """Convert to a Pydantic XML model."""
        return Parameter(id=self.parameter_id, value=self.value)


@dataclass
class UWSJobDescription:
    """Brief job description used for the job list.

    This is a strict subset of the fields of `UWSJob`, but is kept separate
    without an inheritance relationship to reflect how it's used in code.
    """

    job_id: str
    """Unique identifier of the job."""

    message_id: str | None
    """Internal message identifier for the work queuing system."""

    owner: str
    """Identity of the owner of the job."""

    phase: ExecutionPhase
    """Execution phase of the job."""

    run_id: str | None
    """Optional opaque string provided by the client.

    The RunId is intended for the client to add a unique identifier to all
    jobs that are part of a single operation from the perspective of the
    client.  This may aid in tracing issues through a complex system or
    identifying which operation a job is part of.
    """

    creation_time: datetime
    """When the job was created."""

    def to_xml_model(self, base_url: str) -> ShortJobDescription:
        """Convert to a Pydantic XML model.

        Parameters
        ----------
        base_url
            Base URL for the full jobs.
        """
        return ShortJobDescription(
            phase=self.phase,
            run_id=self.run_id,
            creation_time=self.creation_time,
            owner_id=self.owner,
            job_id=self.job_id,
            type=None,
            href=f"{base_url}/{self.job_id}",
        )


class GenericParameters(Parameters):
    """Generic container for UWS job parameters.

    Notes
    -----
    The intended use of `vo_models.uws.Parameters` is to define a subclass
    with the specific parameters valid for that service. However, we store
    parameters as sent to the service as a generic key/value pair in the
    database and define a model that supports arbitrary parsing and
    transformations, so this XML model is both not useful and not clearly
    relevant.

    At least for now, define a generic subclass of `~vo_models.uws.Parameters`
    that holds a generic list of parameters and convert to that for the
    purposes of serialization.
    """

    params: list[Parameter]
    """Job parameters."""


@dataclass
class UWSJob:
    """Represents a single UWS job."""

    job_id: str
    """Unique identifier of the job."""

    message_id: str | None
    """Internal message identifier for the work queuing system."""

    owner: str
    """Identity of the owner of the job."""

    phase: ExecutionPhase
    """Execution phase of the job."""

    run_id: str | None
    """Optional opaque string provided by the client.

    The RunId is intended for the client to add a unique identifier to all
    jobs that are part of a single operation from the perspective of the
    client. This may aid in tracing issues through a complex system or
    identifying which operation a job is part of.
    """

    creation_time: datetime
    """When the job was created."""

    start_time: datetime | None
    """When the job started executing (if it has started)."""

    end_time: datetime | None
    """When the job stopped executing (if it has stopped)."""

    destruction_time: datetime
    """Time at which the job should be destroyed.

    At this time, the job will be aborted if it is still running, its results
    will be deleted, and all record of the job will be discarded.

    This field is optional in the UWS standard, but in this UWS implementation
    all jobs will have a destruction time, so it is not marked as optional.
    """

    execution_duration: timedelta
    """Allowed maximum execution duration in seconds.

    This is specified in elapsed wall clock time, or 0 for unlimited execution
    time. If the job runs for longer than this time period, it will be
    aborted.
    """

    quote: datetime | None
    """Expected completion time of the job if it were started now.

    May be `None` to indicate that the expected duration of the job is not
    known. Maybe later than the destruction time to indicate that the job is
    not possible due to resource constraints.
    """

    error: UWSJobError | None
    """Error information if the job failed."""

    parameters: list[UWSJobParameter]
    """The parameters of the job."""

    results: list[UWSJobResult]
    """The results of the job."""

    def to_xml_model(self) -> JobSummary:
        """Convert to a Pydantic XML model."""
        results = None
        if self.results:
            results = [r.to_xml_model() for r in self.results]
        return JobSummary(
            job_id=self.job_id,
            run_id=self.run_id,
            owner_id=self.owner,
            phase=self.phase,
            quote=self.quote,
            creation_time=self.creation_time,
            start_time=self.start_time,
            end_time=self.end_time,
            execution_duration=int(self.execution_duration.total_seconds()),
            destruction=self.destruction_time,
            parameters=GenericParameters(
                params=[p.to_xml_model() for p in self.parameters]
            ),
            results=results,
            error_summary=self.error.to_xml_model() if self.error else None,
            version=UWSVersion.V1_1,
        )
